@@ -1,6 +1,6 @@
 from gql import gql, Client
 from .gql_client_transport import GqlClientTransport
-from obschart.api.nodes import (
+from .api.nodes import (
     Session,
     Image,
     imageFragment,
@@ -9,13 +9,18 @@ from obschart.api.nodes import (
     programTrackActionResponseFragment,
     ProgramTrackAction,
     programTrackActionFragment,
+    ProgramModule,
+    programModuleFragment,
+    Tag,
+    tagFragment,
 )
-from obschart.api.context import Context
+from .api.context import Context
 from typing import Optional, Dict, Any, Callable
 import datetime
 from .blocks_action_builder import BlocksActionBuilder
 from typing import Union, Awaitable
 import asyncio
+import traceback
 
 
 OnResponseCallback = Callable[[ProgramTrackActionResponse], Awaitable[None]]
@@ -93,11 +98,14 @@ class ObschartClient:
         variables = {"filter": filter}
 
         while True:
-            response = self._execute(query, variables)
-            edges = response["programTrackActionResponses"]["edges"]
+            try:
+                response = self._execute(query, variables)
+                edges = response["programTrackActionResponses"]["edges"]
 
-            if len(edges) > 0:
-                break
+                if len(edges) > 0:
+                    break
+            except Exception as exception:
+                traceback.print_exception(None, exception, exception.__traceback__)
 
             await asyncio.sleep(1)
 
@@ -114,7 +122,15 @@ class ObschartClient:
                     {"applicationId": {"eq": "me"}}
                 )
 
-                await asyncio.gather(*[on_response_callback(response) for response in responses])
+                results = await asyncio.gather(
+                    *[on_response_callback(response) for response in responses],
+                    return_exceptions=True,
+                )
+
+                for result in results:
+                    if isinstance(result, Exception):
+                        exception = result
+                        traceback.print_exception(None, exception, exception.__traceback__)
 
         loop = asyncio.get_event_loop()
         loop.run_until_complete(main())
@@ -265,3 +281,75 @@ class ObschartClient:
 
     def build_feedback_data(self):
         return BlocksActionBuilder()
+
+    async def list_program_modules(self, filter: dict):
+        query = gql(
+            """
+          query ListProgramModules($filter: ProgramModuleFilterInput) {
+            programModules(filter: $filter, first: 9999) {
+              edges {
+                node {
+                  ...ProgramModule
+                }
+              }
+            }
+          }
+        """
+            + programModuleFragment
+        )
+
+        variables = {"filter": filter}
+        response = self._execute(query, variables)
+
+        nodes: list[ProgramModule] = [
+            ProgramModule(edge["node"], self._context)
+            for edge in response["programModules"]["edges"]
+        ]
+
+        return nodes
+
+    async def list_tags(self, filter: dict):
+        query = gql(
+            """
+          query ListTags($filter: TagFilterInput) {
+            tags(filter: $filter, first: 9999) {
+              edges {
+                node {
+                  ...Tag
+                }
+              }
+            }
+          }
+        """
+            + tagFragment
+        )
+
+        variables = {"filter": filter}
+        response = self._execute(query, variables)
+
+        nodes: list[Tag] = [Tag(edge["node"], self._context) for edge in response["tags"]["edges"]]
+
+        return nodes
+
+    def update_program_module(self, id: str, is_public: bool):
+        query = gql(
+            """
+            mutation UpdateProgramModuleMutation($input: UpdateProgramModuleInput!)  {
+                updateProgramModule(input: $input) {
+                    programModule {
+                        ...ProgramModule
+                    }
+                }
+            }
+        """
+            + programModuleFragment
+        )
+
+        input = {
+            "id": id,
+            "isPublic": is_public,
+        }
+        result = self._context.client._execute_mutation(query, input)
+
+        return ProgramModule(result["updateProgramModule"]["programModule"], self._execute,)
+
